@@ -1,7 +1,8 @@
-#include "source-clone.h"
-#include "audio-wrapper.h"
 #include <obs-module.h>
 #include <obs-frontend-api.h>
+#include "util/dstr.h"
+#include "source-clone.h"
+#include "audio-wrapper.h"
 
 const char *source_clone_get_name(void *type_data)
 {
@@ -17,6 +18,7 @@ static void *source_clone_create(obs_data_t *settings, obs_source_t *source)
 	pthread_mutex_init(&context->audio_mutex, NULL);
 	context->cx = 1;
 	context->cy = 1;
+	obs_source_update(source, NULL);
 	return context;
 }
 
@@ -163,6 +165,7 @@ void source_clone_update(void *data, obs_data_t *settings)
 	bool audio_enabled = obs_data_get_bool(settings, "audio");
 	bool active_clone = obs_data_get_bool(settings, "active_clone");
 	context->clone_type = obs_data_get_int(settings, "clone_type");
+	bool async = true;
 	if (context->clone_type == CLONE_SOURCE) {
 		const char *source_name =
 			obs_data_get_string(settings, "clone");
@@ -172,6 +175,8 @@ void source_clone_update(void *data, obs_data_t *settings)
 			source = NULL;
 		}
 		if (source) {
+			async = (obs_source_get_output_flags(source) &
+				 OBS_SOURCE_ASYNC) != 0;
 			if (!obs_weak_source_references_source(context->clone,
 							       source) ||
 			    context->audio_enabled != audio_enabled ||
@@ -188,6 +193,8 @@ void source_clone_update(void *data, obs_data_t *settings)
 	context->num_channels = audio_output_get_channels(obs_get_audio());
 	context->buffer_frame =
 		(uint8_t)obs_data_get_int(settings, "buffer_frame");
+	context->no_filter = obs_data_get_bool(settings, "no_filters") &&
+			     !async;
 }
 
 void source_clone_defaults(obs_data_t *settings)
@@ -210,15 +217,105 @@ bool source_clone_list_add_source(void *data, obs_source_t *source)
 	return true;
 }
 
+struct same_clones {
+	obs_data_t *settings;
+	DARRAY(const char *) clones;
+};
+
+bool find_clones(void *data, obs_source_t *source)
+{
+	if (strcmp(obs_source_get_unversioned_id(source), "source-clone") !=
+	    0) {
+
+		return true;
+	}
+	obs_data_t *settings = obs_source_get_settings(source);
+	if (!settings)
+		return true;
+	struct same_clones *sc = data;
+	if (settings == sc->settings) {
+		obs_data_release(settings);
+		return true;
+	}
+	if (obs_data_get_int(sc->settings, "clone_type") == CLONE_SOURCE) {
+		if (obs_data_get_int(settings, "clone_type") == CLONE_SOURCE &&
+		    strcmp(obs_data_get_string(sc->settings, "clone"),
+			   obs_data_get_string(settings, "clone")) == 0) {
+			const char *name = obs_source_get_name(source);
+			da_push_back(sc->clones, &name);
+		}
+	} else if (obs_data_get_int(sc->settings, "clone_type") ==
+		   obs_data_get_int(settings, "clone_type")) {
+		const char *name = obs_source_get_name(source);
+		da_push_back(sc->clones, &name);
+	}
+	obs_data_release(settings);
+	return true;
+}
+
+void find_same_clones(obs_properties_t *props, obs_data_t *settings)
+{
+	struct same_clones sc;
+	sc.settings = settings;
+	da_init(sc.clones);
+	obs_enum_sources(find_clones, &sc);
+	obs_property_t *prop = obs_properties_get(props, "same_clones");
+	if (sc.clones.num) {
+		struct dstr names;
+		dstr_init_copy(&names, sc.clones.array[0]);
+		for (size_t i = 1; i < sc.clones.num; i++) {
+			dstr_cat(&names, "\n");
+			dstr_cat(&names, sc.clones.array[i]);
+		}
+		obs_data_set_string(settings, "same_clones", names.array);
+		dstr_free(&names);
+		obs_property_set_visible(prop, true);
+	} else {
+		obs_data_unset_user_value(settings, "same_clones");
+		obs_property_set_visible(prop, false);
+	}
+	da_free(sc.clones);
+}
+
+bool source_clone_source_changed(void *priv, obs_properties_t *props,
+				 obs_property_t *property, obs_data_t *settings)
+{
+	UNUSED_PARAMETER(property);
+	struct source_clone *context = priv;
+	const char *source_name = obs_data_get_string(settings, "clone");
+	bool async = true;
+	obs_source_t *source = obs_get_source_by_name(source_name);
+	if (source == context->source) {
+		obs_source_release(source);
+		source = NULL;
+	}
+	if (source) {
+		async = (obs_source_get_output_flags(source) &
+			 OBS_SOURCE_ASYNC) != 0;
+		obs_source_release(source);
+	}
+
+	obs_property_t *no_filters = obs_properties_get(props, "no_filters");
+	obs_property_set_visible(no_filters, !async);
+
+	find_same_clones(props, settings);
+	return true;
+}
+
 bool source_clone_type_changed(void *priv, obs_properties_t *props,
 			       obs_property_t *property, obs_data_t *settings)
 {
 	UNUSED_PARAMETER(priv);
 	UNUSED_PARAMETER(property);
 	obs_property_t *clone = obs_properties_get(props, "clone");
-	obs_property_set_visible(clone,
-				 obs_data_get_int(settings, "clone_type") ==
-					 CLONE_SOURCE);
+	const bool clone_source = obs_data_get_int(settings, "clone_type") ==
+				  CLONE_SOURCE;
+	obs_property_set_visible(clone, clone_source);
+	if (clone_source) {
+		source_clone_source_changed(priv, props, NULL, settings);
+	} else {
+		find_same_clones(props, settings);
+	}
 	return true;
 }
 
@@ -250,6 +347,9 @@ obs_properties_t *source_clone_properties(void *data)
 		source_clone_list_add_source(p, s);
 		obs_source_release(s);
 	}
+	obs_property_set_modified_callback2(p, source_clone_source_changed,
+					    data);
+
 	obs_properties_add_bool(props, "audio", obs_module_text("Audio"));
 	p = obs_properties_add_list(props, "buffer_frame",
 				    obs_module_text("VideoBuffer"),
@@ -263,6 +363,14 @@ obs_properties_t *source_clone_properties(void *data)
 	obs_properties_add_bool(props, "active_clone",
 				obs_module_text("ActiveClone"));
 
+	obs_properties_add_bool(props, "no_filters",
+				obs_module_text("NoFilters"));
+
+	p = obs_properties_add_text(props, "same_clones",
+				    obs_module_text("SameClones"),
+				    OBS_TEXT_INFO);
+	obs_property_set_visible(p, false);
+
 	obs_properties_add_text(
 		props, "plugin_info",
 		"<a href=\"https://obsproject.com/forum/resources/source-clone.1632/\">Source Clone</a> (" PROJECT_VERSION
@@ -364,7 +472,11 @@ void source_clone_video_render(void *data, gs_effect_t *effect)
 		return;
 	}
 	if (context->buffer_frame == 0) {
-		obs_source_video_render(source);
+		if (context->no_filter) {
+			obs_source_default_render(source);
+		} else {
+			obs_source_video_render(source);
+		}
 		obs_source_release(source);
 		context->rendering = false;
 		return;
@@ -404,7 +516,11 @@ void source_clone_video_render(void *data, gs_effect_t *effect)
 		if (context->source_cx && context->source_cy) {
 			gs_ortho(0.0f, (float)context->source_cx, 0.0f,
 				 (float)context->source_cy, -100.0f, 100.0f);
-			obs_source_video_render(source);
+			if (context->no_filter) {
+				obs_source_default_render(source);
+			} else {
+				obs_source_video_render(source);
+			}
 		}
 		gs_texrender_end(context->render);
 
@@ -429,7 +545,8 @@ uint32_t source_clone_get_width(void *data)
 	obs_source_t *source = obs_weak_source_get_source(context->clone);
 	if (!source)
 		return 1;
-	uint32_t width = obs_source_get_width(source);
+	uint32_t width = context->no_filter ? obs_source_get_base_width(source)
+					    : obs_source_get_width(source);
 	obs_source_release(source);
 	if (context->buffer_frame > 1)
 		width /= context->buffer_frame;
@@ -446,7 +563,9 @@ uint32_t source_clone_get_height(void *data)
 	obs_source_t *source = obs_weak_source_get_source(context->clone);
 	if (!source)
 		return 1;
-	uint32_t height = obs_source_get_height(source);
+	uint32_t height = context->no_filter
+				  ? obs_source_get_base_height(source)
+				  : obs_source_get_height(source);
 	obs_source_release(source);
 	if (context->buffer_frame > 1)
 		height /= context->buffer_frame;
@@ -552,8 +671,15 @@ void source_clone_video_tick(void *data, float seconds)
 			obs_source_t *s =
 				obs_weak_source_get_source(context->clone);
 			if (s) {
-				context->source_cx = obs_source_get_width(s);
-				context->source_cy = obs_source_get_height(s);
+				context->source_cx =
+					context->no_filter
+						? obs_source_get_base_width(s)
+						: obs_source_get_width(s);
+				context->source_cy =
+					context->no_filter
+						? obs_source_get_base_height(s)
+						: obs_source_get_height(s);
+
 				cx = context->source_cx;
 				cy = context->source_cy;
 				obs_source_release(s);
@@ -575,7 +701,6 @@ void source_clone_video_tick(void *data, float seconds)
 	if (!context->audio_enabled)
 		return;
 
-
 	const audio_t *a = obs_get_audio();
 	const struct audio_output_info *aoi = audio_output_get_info(a);
 
